
#ifndef SYNTAX_H
#define SYNTAX_H

#define MAX_STORAGE_LENGTH  8000
#define STORAGE_INCREASE    2000

#define MAX_TOKEN_NUM       1000
#define TOKEN_NUM_INCREASE  200

#define MAX_TOKEN_LENGTH  512
#define MAX_LINE_LENGTH   2048

#define IF_NEST_LEVELS 8

#define LINE_CONTINUE_TOKEN "..."
#define COMMENT_TOKEN "%"
#define CARR_RET_TOKEN ";"
#define ODE_LABEL "->"
#define VAR_LABEL ":="
#define ASSIGN_LABEL "="

#define LOOP_TOKEN  "for"
#define IF_TOKEN    "if"
#define THEN_TOKEN  "then"
#define ELSE_TOKEN  "else"
#define ENDIF_TOKEN "endif"

#define ALIAS_NUM 5     /* aliases */

const char aliasArray[2*ALIAS_NUM][10] = 
  { "&&", "and", "||", "or", "Adaptive", "adaptive", "input", "include", "bc_type", "bc.define" };  

#define SPECIAL_NUM 6   /* non-delimited tokens: */

const char specialArray[SPECIAL_NUM][5] = { ODE_LABEL, VAR_LABEL, ASSIGN_LABEL, 
                                            LINE_CONTINUE_TOKEN, COMMENT_TOKEN, CARR_RET_TOKEN };  

//***************************************************************************

inline bool is_letter(const char c) {
  if ( (c >= 'a' && c <='z') || (c >= 'A' && c <='Z') || c == '_' ) return true;
  else return false; }

inline bool is_numeral(const char c) {
  if (c >= '0' && c <= '9') return true;  else return false; }

inline bool is_sign(const char c) {
  if (c == '+' || c == '-') return true;  else return false; }

inline bool is_short_token(const char c) {
  if ( strchr("~!@#%^&*()[]=-+\\|;:/<>?,",c) ) return true;
  else return false;
} 

inline bool is_delim(const char c) {
  if ( !is_letter(c) && !is_numeral(c) && !is_short_token(c) && !strchr(".{}\"\'`$",c) ) return true;
  else return false;
} 

inline char equal(const char *s1, const char *s2) 
   { return ( (strcmp(s1, s2) == 0) && (strlen(s1) == strlen(s2)) ) ? 1 : 0; }

inline char equal_(const char *s1, const char *s2)  // returns "1" if s2 is a prefix of s1
{ return ( strncmp(s2, s1, strlen(s2)) == 0 ); }

inline char *StrCpy(const char *s) 
  { char *d = new char[strlen(s) + 1]; return strcpy(d, s); }

inline char *StrCpy(const char *s1, const char *s2) 
  { char *d = new char[strlen(s1) + strlen(s2) + 2]; strcpy(d,s1); strcat(d,"."); return strcat(d, s2); }

inline char *StrCpy(const char *s, int n) 
  { char *d = new char[n+1]; strncpy(d, s, n); d[n] = '\0'; return d; }

//***************************************************************************


class TokenString
{
  friend class LoopObj;
  friend class InterpolObj;

private:

 long length, max_length;
 char *storage;
 int  max_token_num;
 char **token_ptr;

public:

 int  token_num;

 // constructor and parsing routines:

 TokenString(const char *fname, const char *, int, char **);
 TokenString(const TokenString &);
 ~TokenString()  { delete [] storage; delete [] token_ptr; }

 void parse(const char *fname, const char *extra, int, char **);
 void getToken(char *&ptr0, char *&ptr1, char *, bool);
 void addToken(const char *token);
 void deleteToken(int, int);

 // routines returning the number of occurences of an argument token in the TokenString:

 int token_count(const char *token, int *p=0);
 int token_count(const char *t1, const char *t2, int *p=0) 
    { char *t = ::StrCpy(t1,t2); int n = token_count(t, p); delete [] t; return n; }
 int token2_count(const char *token1, const char *token2, int *p=0);
 int token3_count(const char *token1, const char *token2, const char *token3, int *p=0);
 // int token_count_(const char *token, int *p=0);

 // routines returning the location index of an argument token within the TokenString:

 int token_index(const char *token, int i = 1);
 int token_index(const char *t1, const char *t2, int i = 1)
    { char *t = ::StrCpy(t1,t2); int n = token_index(t, i); delete [] t; return n; }
 int token2_index(const char *token1, const char *token2, int i = 1); 
 int token3_index(const char *token1, const char *token2, const char *token3, int i = 1); 
 // int token_index_(const char *token, int i = 1);

 // routines interpreting the numerical/string value of a token argument:

 bool isParam(const char *s, int *);
 bool isGoodParam(const char *s, double *r = 0);
 bool isConst(const char *s, double *r = 0);

 // routines interpreting the numerical/string value of a token specified by its location:

 double get_double(int p);      
 int    get_int(int p);         
 long   get_long(int p);        

 // routines interpreting the numerical value of a parameter name: 

 double get_param(const char *t, const char *message=0, double *d=0);
 double get2_param(const char *t1, const char *t2, const char *message=0, double *d=0)
    { char *t = ::StrCpy(t1,t2); double x = get_param(t, message, d); delete [] t; return x; }
 int    get_int_param(const char *, const char * = 0, int * = 0);
 long   get_long_param(const char *, const char * = 0, long * = 0);

 // routines interpreting the string value of a token: 

 char  *get_string(const char *);
 char  *get_string(const char *, char *);

 char  *get_string(int p, char *s) { return get_string(token_ptr[p], s); };
 char  *get_string(int p)          { return get_string(token_ptr[p]); };

 char  *line_string(int, char *, class FieldObj *Ca=0, class BufferArray *Bufs=0, class KineticObj *KO=0);
 char  *line_string(int, class FieldObj *Ca=0, class BufferArray *Bufs=0, class KineticObj *KO=0);

 // routines parsing the arguments/variables following a given token on the same line:

 void   trail_pars(const char *token, int m, ... );
 void   trail_pars(const char *token1, const char *token2, int m, ... );

 // routines interpreting the numerical/string value of a token following a token argument:
 
 double trail_double(const char *t, int i=1)    { return get_double( token_index(t, i) + 1 ); }
 int    trail_int(const char *t, int i = 1)     { return get_int( token_index(t, i) + 1 ); }
 long   trail_long(const char *t, int i = 1)    { return get_long( token_index(t, i) + 1 ); }
 char   *trail_string(const char *t, int i = 1) { return get_string( token_index(t, i) + 1 ); }
 char   *trail_string(const char *t, char *s, int i = 1) { return get_string( token_index(t, i) + 1, s); }

 // auxiliary routines:

 int  tokens_to_eol(int p);  // tokens to end of line, with #p included in the count
 int  lastInLine(int p)  { return  p - 1 + tokens_to_eol(p); }

 char equal(int p, const char *s) { return ::equal(token_ptr[p], s); }
 char equal(const char *s, int p) { return ::equal(s, token_ptr[p]); }
 // char equal_(int p, const char *s) { return ::equal_(token_ptr[p], s); }

 char *StrCpy(int p) { return ::StrCpy(token_ptr[p]); } 
 char *operator[](int index) const  { return token_ptr[index]; };
   
 void print(const char *fname);
 void print(FILE * = stderr);
 void linePrint(int p, FILE *f = stderr);
};

#endif
